package nce.editor;

import java.util.ArrayList;
import java.util.List;

import javax.swing.text.Segment;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.rules.IPartitionTokenScanner;
import org.eclipse.jface.text.rules.IToken;
import org.gjt.sp.jedit.Mode;
import org.gjt.sp.jedit.buffer.LineManager;
import org.gjt.sp.jedit.indent.IndentRule;
import org.gjt.sp.jedit.syntax.ModeProvider;
import org.gjt.sp.jedit.syntax.Token;
import org.gjt.sp.jedit.syntax.TokenMarker;
import org.gjt.sp.util.IntegerArray;
import org.gjt.sp.util.Log;

public class ColoringEditorPartitionScanner extends JeditBuffer implements IPartitionTokenScanner {

	private IDocument fDocument;
	private char[] fDocumentText;
	private int fOffset;
	private int fLength;
	private int fPartitionOffset;
	private String fFilename;
	private ColoringTokenHandler fTokenHandler;
	private Token fCurrentToken;
	private Mode fMode;

	public ColoringEditorPartitionScanner(String filename) {
		fFilename = filename;
	}

	@Override
	public void setRange(IDocument document, int offset, int length) {		
		setPartialRange(document, offset, length, null, -1);
	}


	@Override
	public void setPartialRange(IDocument document, int offset, int length, String contentType, int partitionOffset) {
		
		Log.log(Log.DEBUG, this, "setPartialRange: offset="+offset+", length="+length+", partitionOffset="+partitionOffset);

		fOffset = offset;
		fLength = length;
		fPartitionOffset = partitionOffset;

		// optimization does not work ???sync linemanager with changes???
		//if (fDocument == null) { 
			fDocument = document;
			fDocumentText = fDocument.get().toCharArray();
			initJEditBuffer();
		//}
		
		fTokenHandler = null;
	}

	private void initJEditBuffer() {
		
		fMode = loadMode(fDocument, fFilename);
		fTokenMarker = createTokenMarker(fMode);
		fLineMgr = createLineManager(fDocument); 
	}

	@Override
	public IToken nextToken() {

		if (fTokenHandler == null) {
			fTokenHandler = createTokenHandler(fDocument, fOffset, fLength, fPartitionOffset);
			fCurrentToken = findFirstTokenWithinRange(fTokenHandler, fOffset); 						
		} else {
			fCurrentToken = fCurrentToken.next;
		}
		
		if (fCurrentToken == null)
			return ColoringToken.EOF;
		
		ColoringToken result = new ColoringToken(Token.tokenToString(fCurrentToken.id));
		return result;
	}

	private Token findFirstTokenWithinRange(ColoringTokenHandler tokenHandler, int offset) {
		Token token = tokenHandler.getTokens();
		while (token != null && token.offset < offset)
			token = token.next;
		return token;
	}

	private TokenMarker createTokenMarker(Mode mode) {
		
		if (mode != null)
			return mode.getTokenMarker();
			
		Log.log(Log.ERROR, this, "Could not instantiate token marker");
		return null;
		
	}

	public String getModeName() {
		
		if (fMode != null)
			return fMode.getName();
		else
			return null;
	}

	private Mode loadMode(IDocument document, String filename) {
		
		Mode mode = null;
		try {
			
			IRegion firstLineInfo = document.getLineInformation(0);
			String firstLine = document.get(firstLineInfo.getOffset(), firstLineInfo.getLength());
			mode = ModeProvider.instance.getModeForFile(filename, firstLine);
			
		} catch (BadLocationException e) {
			Log.log(Log.ERROR, this, "Could not determine mode for filename '" + filename + "'", e);
		}
		
		return mode;
	}

	private LineManager createLineManager(IDocument document) {
		
		try {
			
			IntegerArray endOffsets = new IntegerArray();

			for (int i = 0; i < document.getNumberOfLines(); i++) {
				int off = document.getLineOffset(i);
				int len= document.getLineLength(i);
				endOffsets.add(off + len);
			}
			
			LineManager lineManager = new LineManager();
			lineManager._contentInserted(endOffsets);
			return lineManager;
			
		} catch (BadLocationException e) {
			Log.log(Log.ERROR, this, "Could not setup line manager", e);
		}
		
		return null;
	}

	@Override
	public int getTokenOffset() {
		return fCurrentToken.offset;
	}

	@Override
	public int getTokenLength() {
		return fCurrentToken.length;
	}

	/*
	 * Implemetation of delegates for IJEditBuffer
	 */
	public int getLineLength(int lineIndex) {
		try {
			return fDocument.getLineLength(lineIndex);
		} catch (BadLocationException e) {
			throw new ArrayIndexOutOfBoundsException(lineIndex);
		}
	}

	public void getLineText(int line, Segment lineText) {

		if(line < 0 || line >= fDocument.getNumberOfLines())
			throw new ArrayIndexOutOfBoundsException(line);
		
		try {
			IRegion lineInfo = fDocument.getLineInformation(line);
			lineText.array = fDocumentText;
			lineText.offset = lineInfo.getOffset();
			lineText.count = lineInfo.getLength();
		} catch (BadLocationException e) {
			throw new ArrayIndexOutOfBoundsException(line);
		}
	}

	public int getLineCount() {
		return fDocument.getNumberOfLines();
	}

	public int getLineStartOffset(int line) {
		try {
			return fDocument.getLineOffset(line);
		} catch (BadLocationException e) {
			throw new ArrayIndexOutOfBoundsException(line);
		}
	}

	public int getLineOfOffset(int openBracketIndex) {
		try {
			return fDocument.getLineOfOffset(openBracketIndex);
		} catch (BadLocationException e) {
			throw new ArrayIndexOutOfBoundsException(openBracketIndex);
		}
	}

	public CharSequence getLineSegment(int index) {
		try {
			IRegion region = fDocument.getLineInformation(index);
			return fDocument.get(region.getOffset(), region.getLength());
		} catch (BadLocationException e) {
			throw new ArrayIndexOutOfBoundsException(index);
		}
	}

	public List<IndentRule> getIndentRules() {
		if (fMode != null)
			return fMode.getIndentRules();
		else 
			return new ArrayList<IndentRule>();
	}

	@Override
	public Mode getMode() {
		return fMode;
	}

}
