package nce.editor;

import java.util.LinkedList;
import java.util.List;

import javax.swing.text.Segment;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.ui.editors.text.EditorsUI;
import org.eclipse.ui.texteditor.AbstractDecoratedTextEditorPreferenceConstants;
import org.gjt.sp.jedit.Debug;
import org.gjt.sp.jedit.Mode;
import org.gjt.sp.jedit.buffer.IJEditBuffer;
import org.gjt.sp.jedit.buffer.LineManager;
import org.gjt.sp.jedit.indent.IndentAction;
import org.gjt.sp.jedit.indent.IndentRule;
import org.gjt.sp.jedit.syntax.DummyTokenHandler;
import org.gjt.sp.jedit.syntax.ModeProvider;
import org.gjt.sp.jedit.syntax.TokenHandler;
import org.gjt.sp.jedit.syntax.TokenMarker;
import org.gjt.sp.jedit.syntax.TokenMarker.LineContext;
import org.gjt.sp.util.Log;
import org.gjt.sp.util.StandardUtilities;

public abstract class JeditBuffer implements IJEditBuffer {

	protected LineManager fLineMgr = null;

	private boolean fContextInsensitive = false;

	protected TokenMarker fTokenMarker = null;

	public abstract int getLineLength(int lineIndex);

	public abstract Mode getMode();

	/*
	 * Copied from JEditBuffer.
	 */
	@Override
	public void markTokens(int lineIndex, TokenHandler tokenHandler) {

		Segment seg = new Segment();

		if(lineIndex < 0 || lineIndex >= fLineMgr .getLineCount())
			throw new ArrayIndexOutOfBoundsException(lineIndex);

		int firstInvalidLineContext = fLineMgr.getFirstInvalidLineContext();
		int start;
		if(fContextInsensitive  || firstInvalidLineContext == -1)
		{
			start = lineIndex;
		}
		else
		{
			start = Math.min(firstInvalidLineContext,
				lineIndex);
		}

		if(Debug.TOKEN_MARKER_DEBUG)
			Log.log(Log.DEBUG,this,"tokenize from " + start + " to " + lineIndex);
		
		TokenMarker.LineContext oldContext = null;
		TokenMarker.LineContext context = null;
		for(int i = start; i <= lineIndex; i++)
		{
			getLineText(i,seg);

			oldContext = fLineMgr.getLineContext(i);

			TokenMarker.LineContext prevContext = (
				(i == 0 || fContextInsensitive) ? null
				: fLineMgr.getLineContext(i - 1)
			);

			TokenHandler _tokenHandler = i == lineIndex ? tokenHandler : DummyTokenHandler.INSTANCE;
			context = markTokens(seg, prevContext, _tokenHandler);
			fLineMgr.setLineContext(i,context);
		}

		int lineCount = fLineMgr.getLineCount();
		if(lineCount - 1 == lineIndex)
			fLineMgr.setFirstInvalidLineContext(-1);
		else if(oldContext != context)
			fLineMgr.setFirstInvalidLineContext(lineIndex + 1);
		else if(firstInvalidLineContext == -1)
			/* do nothing */;
		else
		{
			fLineMgr.setFirstInvalidLineContext(Math.max(firstInvalidLineContext,lineIndex + 1));
		}
	}

	private LineContext markTokens(Segment seg, LineContext prevContext, TokenHandler _tokenHandler) {
		TokenMarker.LineContext context = fTokenMarker.markTokens(prevContext, _tokenHandler, seg);
		return context;
	}

	@Override
	public int getTabSize() {
		return EditorsUI.getPreferenceStore().getInt(AbstractDecoratedTextEditorPreferenceConstants.EDITOR_TAB_WIDTH);
	}

	@Override
	public int getIndentSize() {
		return getTabSize();
	}

	public abstract CharSequence getLineSegment(int index);

	public abstract void getLineText(int line, Segment lineText);

	@Override
	public abstract int getLineStartOffset(int line);

	@Override
	public abstract int getLineCount();

	@Override
	public abstract int getLineOfOffset(int openBracketIndex);

	@Override
	public boolean getBooleanProperty(String string) {
		// TODO replace by implementation
		// is used by indentation to get property "multipleBracketIndent"
		return false;
	}

	protected ColoringTokenHandler createTokenHandler(IDocument document, int offset, int length, int partitionOffset) {
				
		try {
			
			ColoringTokenHandler tokenHandler = new ColoringTokenHandler();
			int startOffset = partitionOffset == -1 ? offset : partitionOffset;
			int firstLineToScan = document.getLineOfOffset(startOffset);
			int lastLineToScan = document.getLineOfOffset(length + offset);
			
			for (int lineIndex = firstLineToScan; lineIndex <= lastLineToScan; lineIndex++)
				markTokens(lineIndex, tokenHandler);
			
			return tokenHandler;
			
		} catch (BadLocationException e) {
			Log.log(Log.ERROR, this, "Could not instantiate token marker", e);
		}
		
		return null;
	}

	/**
	 * Returns the ideal leading indent for the specified line.
	 * This will apply the various auto-indent rules.
	 * @param lineIndex The line number
	 */
	public int getIdealIndentForLine(int lineIndex)
	{
		int prevLineIndex = getPriorNonEmptyLine(lineIndex);
		int oldIndent = prevLineIndex == -1 
							? 0 
							: StandardUtilities.getLeadingWhiteSpaceWidth(getLineSegment(prevLineIndex), getTabSize());
		
		return getIdealIndentForLine(lineIndex, prevLineIndex, oldIndent);
	}

	/**
	 * Returns the ideal leading indent for the specified line.
	 * This will apply the various auto-indent rules.
	 * @param lineIndex The line number
	 * @param prevLineIndex The index of the previous non-empty line
	 * @param oldIndent The indent width of the previous line (or 0)
	 */
	private int getIdealIndentForLine(int lineIndex, int prevLineIndex, int oldIndent)
	{
		int prevPrevLineIndex = prevLineIndex < 0 ? -1 
												  : getPriorNonEmptyLine(prevLineIndex);
		int newIndent = oldIndent;

		List<IndentRule> indentRules = getIndentRules(lineIndex);
		List<IndentAction> actions = new LinkedList<IndentAction>();
		for (int i = 0; i<indentRules.size();  i++)
		{
			IndentRule rule = indentRules.get(i);
			rule.apply(this, lineIndex, prevLineIndex, prevPrevLineIndex, actions);
		}


		for (IndentAction action : actions)
		{
			newIndent = action.calculateIndent(this, lineIndex, oldIndent, newIndent);
			if (!action.keepChecking())
				break;
		}
		if (newIndent < 0)
			newIndent = 0;

		return newIndent;
	}

	/**
	 * Auto indent needs this.
	 */
	public int getPriorNonEmptyLine(int lineIndex)
	{
		if (getMode() != null && !getMode().getIgnoreWhitespace())
		{
			return lineIndex - 1;
		}

		int returnValue = -1;
		for(int i = lineIndex - 1; i >= 0; i--)
		{
			Segment seg = new Segment();
			getLineText(i,seg);
			
			if(seg.count != 0)
				returnValue = i;
			
			for(int j = 0; j < seg.count; j++)
			{
				char ch = seg.array[seg.offset + j];
				if(!Character.isWhitespace(ch))
					return i;
			}
		}

		// didn't find a line that contains non-whitespace chars
		// so return index of prior whitespace line
		return returnValue;
	}

	private List<IndentRule> getIndentRules(int line)
	{
		String modeName = null;
		TokenMarker.LineContext ctx = fLineMgr.getLineContext(line);
		if (ctx != null && ctx.rules != null)
			modeName = ctx.rules.getModeName();
		if (modeName == null)
			modeName = fTokenMarker.getMainRuleSet().getModeName();
		return ModeProvider.instance.getMode(modeName).getIndentRules();
	}

	


	//{{{ getCurrentIndentForLine() method
	/**
	 * Returns the line's current leading indent.
	 * @param lineIndex The line number
	 * @param whitespaceChars If this is non-null, the number of whitespace
	 * characters is stored at the 0 index
	 * @since jEdit 4.2pre2
	 */
	public int getCurrentIndentForLine(int lineIndex, int[] whitespaceChars)
	{
		Segment seg = new Segment();
		getLineText(lineIndex,seg);

		int tabSize = getTabSize();

		int currentIndent = 0;
loop:	for(int i = 0; i < seg.count; i++)
		{
			char c = seg.array[seg.offset + i];
			switch(c)
			{
			case ' ':
				currentIndent++;
				if(whitespaceChars != null)
					whitespaceChars[0]++;
				break;
			case '\t':
				currentIndent += tabSize - (currentIndent % tabSize);
				if(whitespaceChars != null)
					whitespaceChars[0]++;
				break;
			default:
				break loop;
			}
		}

		return currentIndent;
	} //}}}

}
